import sympy as sp
import networkx as nx
from functools import cache
from ..core.structure import create_matrix
from ..core.stability import system_feedback, net_feedback, absolute_feedback
from ..core.helper import get_nodes, get_weight

@cache
def structural_sensitivity(G, level=None):
    A = create_matrix(G, "signed")
    n = A.shape[0]
    fcp = system_feedback(G)[1:]
    if level is None:
        level = n
    S = sp.zeros(n, n)
    nodes = get_nodes(G, "state")
    for i in range(n):
        for j in range(n):
            if A[i, j] != 0:
                sG = nx.DiGraph(G)
                sG[nodes[j]][nodes[i]]["sign"] = 0
                scp = system_feedback(sG)[1:]
                if level <= len(fcp) and level <= len(scp):
                    N = fcp[level - 1] - scp[level - 1]
                    S[i, j] = N
    return S

@cache
def net_structural_sensitivity(G, level=None):
    A = create_matrix(G, "signed")
    n = A.shape[0]
    fcp = net_feedback(G)[1:]
    if level is None:
        level = n
    S = sp.zeros(n, n)
    nodes = get_nodes(G, "state")
    for i in range(n):
        for j in range(n):
            if A[i, j] != 0:
                sG = nx.DiGraph(G)
                sG[nodes[j]][nodes[i]]["sign"] = 0
                scp = net_feedback(sG)[1:]
                if level <= len(fcp) and level <= len(scp):
                    N = fcp[level - 1] - scp[level - 1]
                    S[i, j] = N
    return S


@cache
def absolute_structural_sensitivity(G, level=None):
    A = create_matrix(G, "signed")
    n = A.shape[0]
    fcp = absolute_feedback(G)[1:]
    if level is None:
        level = n
    S = sp.zeros(n, n)
    nodes = get_nodes(G, "state")
    for i in range(n):
        for j in range(n):
            if A[i, j] != 0:
                sG = nx.DiGraph(G)
                sG[nodes[j]][nodes[i]]["sign"] = 0
                scp = absolute_feedback(sG)[1:]
                if level <= len(fcp) and level <= len(scp):
                    N = fcp[level - 1] - scp[level - 1]
                    S[i, j] = N
    return S


@cache
def weighted_structural_sensitivity(G, level=None):
    A = create_matrix(G, "signed")
    n = A.shape[0]
    if level is None:
        level = n
    net = sp.Matrix(net_structural_sensitivity(G, level))
    absolute = sp.Matrix(absolute_structural_sensitivity(G, level))
    return get_weight(net, absolute)



import numpy as np
import sympy as sp
from functools import cache
from .structure import create_matrix
from .helper import perm, get_weight, get_nodes, sign_determinacy


@cache
def adjoint_matrix(G, form="symbolic", perturb=None):
    A = create_matrix(G, form=form)
    A = sp.Matrix(-A)
    nodes = get_nodes(G, "state")
    n = len(nodes)
    if perturb is not None:
        src_id = nodes.index(perturb)
        return sp.Matrix(
            [sp.Integer(-1) ** (src_id + j) * A.minor(src_id, j) 
             for j in range(n)]
        )
    adjoint_matrix = sp.expand(A.adjugate())
    return sp.Matrix(adjoint_matrix)


@cache
def absolute_feedback_matrix(G, perturb=None):
    A = create_matrix(G, form="binary")
    A_np = np.array(sp.matrix2numpy(A), dtype=int)
    nodes = get_nodes(G, "state")
    n = A_np.shape[0]
    if perturb is not None:
        perturb_index = nodes.index(perturb)
        result = np.zeros(n, dtype=int)
        for j in range(n):
            minor = np.delete(np.delete(A_np, perturb_index, 0), j, 1)
            result[j] = int(perm(minor.astype(float)))
        return sp.Matrix(result)
    tmat = np.zeros((n, n), dtype=int)
    for i in range(n):
        for j in range(n):
            minor = np.delete(np.delete(A_np, j, 0), i, 1)
            tmat[i, j] = int(perm(minor.astype(float)))
    return sp.Matrix(tmat)


@cache
def weighted_predictions_matrix(G, perturb=None, as_nan=True, as_abs=False):
    amat = adjoint_matrix(G, perturb=perturb, form="signed")
    if as_abs:
        amat = sp.Abs(amat)
    tmat = absolute_feedback_matrix(G, perturb=perturb)
    if as_nan:
        wmat = get_weight(amat, tmat)
    else:
        wmat = get_weight(amat, tmat, sp.Integer(1))
    return sp.Matrix(wmat)


@cache
def sign_determinacy_matrix(
    G, perturb=None, method="average", as_nan=True, as_abs=False
):
    wmat = weighted_predictions_matrix(
        G, perturb=perturb, as_nan=as_nan, as_abs=as_abs
        )
    tmat = sp.Matrix(absolute_feedback_matrix(G, perturb=perturb))
    pmat = sign_determinacy(wmat, tmat, method)
    return sp.Matrix(pmat)


@cache
def numerical_simulations(G, perturb=None, n_sim=10000, dist="uniform", seed=42, as_abs=False):
    np.random.seed(seed)
    A = create_matrix(G, form="symbolic", matrix_type="A")
    state_nodes = get_nodes(G, "state")
    node_idx = {node: i for i, node in enumerate(state_nodes)}
    n = len(state_nodes)
    symbols = list(A.free_symbols)
    A_sp = sp.lambdify(symbols, A)
    pert_idx, sign = (
        (node_idx[perturb[0]], perturb[1]) 
        if perturb else (None, 1)
    )
    dist_funcs = {
        "uniform": lambda size: np.random.uniform(0, 1, size),
        "weak": lambda size: np.random.beta(1, 3, size),
        "moderate": lambda size: np.random.beta(2, 2, size),
        "strong": lambda size: np.random.beta(3, 1, size),
    }
    positive = np.zeros((n, n), dtype=int)
    negative = np.zeros((n, n), dtype=int)
    total_simulations = 0
    while total_simulations < n_sim:
        values = dist_funcs[dist](len(symbols))
        sim_A = A_sp(*values)
        if np.all(np.real(np.linalg.eigvals(sim_A)) < 0):
            try:
                inv_A = np.linalg.inv(-sim_A)
                effect = (
                    inv_A[:, pert_idx] * sign 
                    if pert_idx is not None else inv_A
                )
                positive += effect > 0
                negative += effect < 0
                total_simulations += 1
            except np.linalg.LinAlgError:
                continue
    smat = np.where(negative > positive, -negative / n_sim, positive / n_sim)
    smat = sp.Matrix(smat.astype(float).tolist())
    tmat = absolute_feedback_matrix(G)
    tmat_np = np.array(tmat.tolist(), dtype=bool)
    smat = sp.Matrix(
        [
            [sp.nan if not tmat_np[i, j] else smat[i, j] for j in range(n)]
            for i in range(n)
        ]
    )
    if as_abs:
        smat = sp.Abs(smat)
    return sp.Matrix(smat)



import pandas as pd
import networkx as nx
import sympy as sp
from functools import cache
from ..core.structure import create_matrix
from ..core.stability import system_feedback, net_feedback, absolute_feedback, weighted_feedback
from ..core.helper import get_nodes, get_weight, get_positive, get_negative, sign_string, arrows

@cache
def get_paths(G, source=None, target=None, form="symbolic"):
    nodes = get_nodes(G, "state")
    A = create_matrix(G, form=form)

    def calculate_paths(s, t):
        if s not in nodes or t not in nodes:
            raise ValueError("Source or target node not found in the graph")
        path_nodes = list(nx.all_simple_paths(G, s, t))
        if s == t:
            return [A[nodes.index(s), nodes.index(t)]]
        elif not path_nodes:
            return [sp.Integer(0)]
        return [
            sp.prod(
                A[nodes.index(p[i + 1]), nodes.index(p[i])] for i in range(len(p) - 1)
            )
            for p in path_nodes
        ]

    if source is None and target is None:
        return sp.Matrix([[sum(calculate_paths(s, t)) for s in nodes] for t in nodes])
    elif source is not None and target is not None:
        if source not in nodes or target not in nodes:
            raise ValueError("Source or target node not found in the graph")
        if source == target:
            return sp.Matrix([A[nodes.index(source), nodes.index(target)]])
        elif not nx.has_path(G, source, target):
            return sp.Matrix([sp.Integer(0)])
        else:
            return sp.Matrix(calculate_paths(source, target))
    else:
        raise ValueError("Both source and target must be specified.")


def paths_table(G, source=None, target=None, form="symbolic"):
    nodes = get_nodes(G, "state")

    def calculate_paths(s, t):
        if s not in nodes or t not in nodes:
            raise ValueError("Source or target node not found in the graph")
        if s == t:
            return [[s, s]] if G.has_edge(s, s) else None
        if not nx.has_path(G, s, t):
            return None
        paths = list(nx.all_simple_paths(G, s, t))
        return paths

    all_paths = []
    if source is None and target is None:
        for s in nodes:
            for t in nodes:
                paths = calculate_paths(s, t)
                if paths is not None:
                    all_paths.extend(paths)
    elif source is not None and target is not None:
        all_paths = calculate_paths(source, target)
    else:
        raise ValueError("Both source and target must be specified.")
    if not all_paths:
        return None
    net_fb = complementary_feedback(G, source=source, target=target, form="signed")
    absolute_fb = complementary_feedback(G, source=source, target=target, form="binary")

    positive_fb = get_positive(net_fb, absolute_fb)
    negative_fb = get_negative(net_fb, absolute_fb)
    weighted_fb = get_weight(net_fb, absolute_fb, 0)
    prediction_weight = weighted_paths(G, source, target, input="positive")

    paths_df = pd.DataFrame(
        {
            "Length": [len(path) - 1 for path in all_paths],
            "Path": [arrows(G, path) for path in all_paths],
            "Path sign": [sign_string(G, path) for path in all_paths],
            "Positive feedback": [positive_fb[i] for i in range(len(all_paths))],
            "Negative feedback": [negative_fb[i] for i in range(len(all_paths))],
            "Weighted feedback": [weighted_fb[i] for i in range(len(all_paths))],
            "Weighted path": [prediction_weight[i] for i in range(len(all_paths))],
        }
    )
    return paths_df.sort_values(["Length", "Path"]).reset_index(drop=True)


def get_cycles(G):
    A = create_matrix(G, form="symbolic")
    nodes = get_nodes(G, "state")
    node_id = {n: i for i, n in enumerate(nodes)}
    cycle_list = nx.simple_cycles(G)
    cycle_nodes = sorted([c for c in cycle_list], key=lambda x: len(x))
    C = [c + [c[0]] for c in cycle_nodes]
    cycles = sp.Matrix(
        [
            sp.prod([A[node_id[c[i + 1]], node_id[c[i]]] for i in range(len(c) - 1)])
            for c in C
        ]
    )
    return cycles


def cycles_table(G):
    cycle_nodes = sorted(
        [path for path in nx.simple_cycles(G)], key=lambda x: (len(x), x)
    )
    all_cycles = [cycle + [cycle[0]] for cycle in cycle_nodes]
    cycle_signs = [sign_string(G, path) for path in all_cycles]
    cycles_df = pd.DataFrame(
        {
            "Length": [len(nodes) for nodes in cycle_nodes],
            "Cycle": [arrows(G, path) for path in all_cycles],
            "Sign": cycle_signs,
        }
    )
    return cycles_df


@cache
def complementary_feedback(G, source=None, target=None, form="symbolic"):
    nodes = get_nodes(G, "state")
    n = len(nodes)

    def calculate_feedback(s, t):
        path_nodes = list(nx.all_simple_paths(G, s, t))
        feedback = []
        for path in path_nodes:
            path_nodes_set = set(path)
            subsystem_nodes = [node for node in nodes if node not in path_nodes_set]
            subsystem = G.subgraph(subsystem_nodes).copy()
            level = n - len(path)
            if form == "symbolic":
                feedback.append(system_feedback(subsystem, level=level)[0])
            elif form == "signed":
                feedback.append(net_feedback(subsystem, level=level)[0])
            elif form == "binary":
                feedback.append(absolute_feedback(subsystem, level=level)[0])
            else:
                raise ValueError(
                    "Invalid form. Choose 'symbolic', 'signed', or 'binary'."
                )
        return [sp.expand_mul(f) for f in feedback]

    if source is None and target is None:
        return sp.Matrix(
            [[sum(calculate_feedback(s, t)) for s in nodes] for t in nodes]
        )
    elif source is not None and target is not None:
        return sp.Matrix(calculate_feedback(source, target))
    else:
        raise ValueError(
            "Both source and target must be specified, or both must be None."
        )


@cache
def system_paths(G, source=None, target=None, input="positive", form="symbolic"):
    def calculate_effect(s, t):
        if s == t:
            feedback = complementary_feedback(G, s, t, form=form)
            if form == "binary":
                return feedback
            else:
                return feedback / sp.Integer(-1)
        else:
            path = get_paths(G, s, t, form=form)
            feedback = complementary_feedback(G, s, t, form=form)
            if path.shape == (0, 0) or feedback.shape == (0, 0):
                return sp.Matrix([[0]])
            if form == "binary":
                effect = path.multiply_elementwise(feedback)
            else:
                effect = path.multiply_elementwise(feedback) / sp.Integer(-1)
            input_multiplier = sp.Integer(1) if input == "positive" else sp.Integer(-1)
            effect = effect.applyfunc(lambda x: x * input_multiplier)
            return sp.Matrix([sp.expand_mul(e) for e in effect])

    nodes = get_nodes(G, "state")
    if source is None and target is None:
        return sp.Matrix([[sum(calculate_effect(s, t)) for s in nodes] for t in nodes])
    elif source is not None and target is not None:
        return calculate_effect(source, target)
    else:
        raise ValueError(
            "Both source and target must be specified, or both must be None."
        )


@cache
def weighted_paths(G, source, target, input="positive"):
    nodes = get_nodes(G, "state")
    A_sgn = create_matrix(G, form="signed")
    input_multiplier = sp.Integer(1) if input == "positive" else sp.Integer(-1)

    def calculate_weighted_paths(s, t):
        path_nodes = list(nx.all_simple_paths(G, s, t))
        wgt_effects = []
        for path in path_nodes:
            subsystem_nodes = [node for node in nodes if node not in path]
            if not subsystem_nodes:
                feedback = sp.Integer(-1)
            else:
                subsystem = G.subgraph(subsystem_nodes).copy()
                feedback = weighted_feedback(subsystem, level=len(nodes) - len(path))
                if feedback[0] == sp.nan:
                    feedback = sp.Integer(0)
            if s == t:
                wgt_effect = sp.Integer(-1) * feedback
            else:
                sign = sp.prod(
                    A_sgn[nodes.index(path[i + 1]), nodes.index(path[i])]
                    for i in range(len(path) - 1)
                )
                wgt_effect = sp.Integer(-1) * sign * feedback * input_multiplier
            wgt_effects.append(wgt_effect)
        return sp.Matrix(wgt_effects)

    return calculate_weighted_paths(source, target)



import numpy as np
import sympy as sp
import networkx as nx
from numba import jit

def list_to_digraph(matrix, ids=None):
    if not isinstance(matrix, (list, np.ndarray)):
        raise ValueError("Input must be a list of lists or a numpy array")
    if isinstance(matrix, list):
        matrix = np.array(matrix)
    if matrix.ndim != 2 or matrix.shape[0] != matrix.shape[1]:
        raise ValueError("Input must be a square matrix")
    G = nx.DiGraph()
    n = matrix.shape[0]
    if ids is None:
        node_ids = [str(i) for i in range(1, n + 1)]
    else:
        if len(ids) != n:
            raise ValueError("Number of ids must match matrix dimensions")
        node_ids = ids
    G.add_nodes_from(node_ids)
    for i in range(n):
        for j in range(n):
            if matrix[i][j] != 0:
                G.add_edge(node_ids[j], node_ids[i], sign=int(matrix[i][j]))
    nx.set_node_attributes(G, "state", "category")
    return G


def digraph_to_list(G):
    if not isinstance(G, nx.DiGraph):
        raise TypeError("Input must be a networkx.DiGraph.")
    n = G.number_of_nodes()
    nodes = sorted(G.nodes())
    node_to_index = {node: i for i, node in enumerate(nodes)}
    matrix = [[0 for _ in range(n)] for _ in range(n)]
    for source, target, data in G.edges(data=True):
        i, j = node_to_index[source], node_to_index[target]
        sign = data.get("sign", 1)
        matrix[j][i] = sign
    return str(matrix)


def powerplay_labels(input_str):
    return [item.split(": ")[1] for item in input_str.split(", ")]


def perm(A, method="glynn"):
    if not isinstance(A, np.ndarray):
        raise TypeError("Input matrix must be a NumPy array.")
    matshape = A.shape
    if matshape[0] != matshape[1]:
        raise ValueError("Input matrix must be square.")
    if np.isnan(A).any():
        raise ValueError("Input matrix must not contain NaNs.")
    if matshape[0] == 0:
        return 1
    if matshape[0] == 1:
        return A[0, 0]
    if matshape[0] == 2:
        return A[0, 0] * A[1, 1] + A[0, 1] * A[1, 0]
    if matshape[0] == 3:
        return (
            A[0, 2] * A[1, 1] * A[2, 0]
            + A[0, 1] * A[1, 2] * A[2, 0]
            + A[0, 2] * A[1, 0] * A[2, 1]
            + A[0, 0] * A[1, 2] * A[2, 1]
            + A[0, 1] * A[1, 0] * A[2, 2]
            + A[0, 0] * A[1, 1] * A[2, 2]
        )
    return _ryser(A) if method != "glynn" else _glynn(A)


@jit(nopython=True)
def _ryser(A):
    n = len(A)
    if n == 0:
        return 1
    row_comb = np.zeros((n), dtype=A.dtype)
    total = 0
    old_grey = 0
    sign = +1
    binary_power_dict = [2**i for i in range(n)]
    num_loops = 2**n
    for k in range(0, num_loops):
        bin_index = (k + 1) % num_loops
        reduced = np.prod(row_comb)
        total += sign * reduced
        new_grey = bin_index ^ (bin_index // 2)
        grey_diff = old_grey ^ new_grey
        grey_diff_index = binary_power_dict.index(grey_diff)
        new_vector = A[grey_diff_index]
        direction = (old_grey > new_grey) - (old_grey < new_grey)
        for i in range(n):
            row_comb[i] += new_vector[i] * direction
        sign = -sign
        old_grey = new_grey
    return total


@jit(nopython=True)
def _glynn(A):
    n = len(A)
    if n == 0:
        return 1
    row_comb = np.sum(A, 0)
    total = 0
    old_gray = 0
    sign = +1
    binary_power_dict = [2**i for i in range(n)]
    num_loops = 2 ** (n - 1)
    for bin_index in range(1, num_loops + 1):
        reduced = np.prod(row_comb)
        total += sign * reduced
        new_gray = bin_index ^ (bin_index // 2)
        gray_diff = old_gray ^ new_gray
        gray_diff_index = binary_power_dict.index(gray_diff)
        new_vector = A[gray_diff_index]
        direction = 2 * ((old_gray > new_gray) - (old_gray < new_gray))
        for i in range(n):
            row_comb[i] += new_vector[i] * direction
        sign = -sign
        old_gray = new_gray
    return total / num_loops


def get_nodes(G, node_type="state", labels=False):
    if not isinstance(G, nx.DiGraph):
        raise TypeError("Input must be a networkx.DiGraph.")

    if node_type == "all":
        return list(G.nodes()) if not labels else list(G.nodes(data=True))
    else:
        return [
            n if not labels else d.get("label", n)
            for n, d in G.nodes(data=True)
            if d.get("category") == node_type
        ]


def get_weight(net, absolute, no_effect=sp.nan):
    if net.shape != absolute.shape:
        raise ValueError("Matrices must have the same shape")
    result = sp.zeros(*net.shape)
    for i in range(net.shape[0]):
        for j in range(net.shape[1]):
            if absolute[i, j] == 0:
                result[i, j] = no_effect
            else:
                result[i, j] = net[i, j] / absolute[i, j]
    return result


def get_positive(net, absolute):
    if net.shape != absolute.shape:
        raise ValueError("Matrices must have the same shape")
    result = sp.zeros(*net.shape)
    for i in range(net.shape[0]):
        for j in range(net.shape[1]):
            result[i, j] = (net[i, j] + absolute[i, j]) // 2
    return result


def get_negative(net, absolute):
    if net.shape != absolute.shape:
        raise ValueError("Matrices must have the same shape")
    result = sp.zeros(*net.shape)
    for i in range(net.shape[0]):
        for j in range(net.shape[1]):
            result[i, j] = (absolute[i, j] - net[i, j]) // 2
    return result


def sign_determinacy(wmat, tmat, method="average"):
    def compute_prob(w, t, method):
        if w == sp.Integer(0):
            return sp.Rational(1, 2)
        elif w == sp.Integer(1):
            return sp.Integer(1)
        elif w == sp.Integer(-1):
            return sp.Integer(-1)
        elif t == sp.Integer(0):
            return sp.nan
        return (
            compute_prob_average(w, t)
            if method == "average"
            else compute_prob_95_bound(w, t)
        )

    def compute_prob_average(w, t):
        bw = 3.45962
        bwt = 0.03417
        prob = sp.exp(bw * w + bwt * w * t) / (1 + sp.exp(bw * w + bwt * w * t))
        return max(sp.Rational(1, 2), prob)

    def compute_prob_95_bound(w, t):
        bw = 9.766
        bwt = 0.139
        prob = sp.exp(bw * w + bwt * w * t) / (1253.992 + sp.exp(bw * w + bwt * w * t))
        return max(sp.Rational(1, 2), prob)

    if method not in ["average", "95_bound"]:
        raise ValueError("Invalid method. Choose 'average' or '95_bound'.")

    rows, cols = wmat.shape

    def calc_prob(i, j):
        w, t = wmat[i, j], tmat[i, j]
        if w.is_zero:
            return sp.Rational(1, 2)
        prob = compute_prob(sp.Abs(w), t, method)
        return sp.sign(w) * prob if prob is not None else sp.nan

    pmat = sp.Matrix(rows, cols, lambda i, j: calc_prob(i, j))
    return pmat


def arrows(G, path):
    arrows = []
    for i in range(len(path) - 1):
        if G[path[i]][path[i + 1]]["sign"] > 0:
            arrows.append(f"{path[i]} $\\rightarrow$")  # Right arrow
        else:
            arrows.append(f"{path[i]} $\\multimap$")  # Multimap
    arrows.append(str(path[-1]))
    return " ".join(arrows)


def sign_string(G, path):
    signs = []
    for from_node, to_node in zip(path, path[1:]):
        sign = G[from_node][to_node]["sign"]
        if sign != 0:
            signs.append(int(sign))
    product = sp.prod(signs)
    if product > 0:
        return "+"
    elif product < 0:
        return "\u2212"
    else:
        return "0"


def symbolic_path(G, A, path, nodes):
    if len(path) == 1:
        return str(A[nodes.index(path[0]), nodes.index(path[0])])
    return " * ".join(
        str(A[nodes.index(path[i]), nodes.index(path[i + 1])])
        for i in range(len(path) - 1)
    )




import json
import networkx as nx
import sympy as sp
from .helper import get_nodes


def import_digraph(data, file_path=True):
    if file_path:
        with open(data, "r") as file:
            data = json.load(file)
    G = nx.DiGraph()
    for node in data["nodes"]:
        att = {k: v for k, v in node.items() if k != "id"}
        G.add_node(node["id"], **att)
    for edge in data["edges"]:
        source, target = edge["from"], edge["to"]
        att = {k: v for k, v in edge.items() if k not in ["from", "to", "arrows"]}
        arr = edge.get("arrows", {}).get("to", {})
        if isinstance(arr, dict):
            arr_type = arr.get("type")
            if arr_type == "triangle":
                att["sign"] = 1
            elif arr_type == "circle":
                att["sign"] = -1
        G.add_edge(source, target, **att)
    nx.set_node_attributes(G, "state", "category")
    return G


def create_matrix(G, form="symbolic", matrix_type="A"):
    if not isinstance(G, nx.DiGraph):
        raise TypeError("Input must be a networkx.DiGraph.")
    if form not in ["symbolic", "signed", "binary"]:
        raise ValueError("Invalid form. Choose 'symbolic', 'signed', or 'binary'.")
    if matrix_type not in ["A", "B", "C", "D"]:
        raise ValueError("Invalid matrix_type. Choose 'A', 'B', 'C', or 'D'.")

    def sym(source, target, prefix):
        return sp.Symbol(f"{prefix}_{target},{source}")

    def sign(source, target, prefix):
        if form == "symbolic":
            return sym(source, target, prefix) * G[source][target].get("sign", 1)
        elif form == "signed":
            return G[source][target].get("sign", 1)
        else:  # form == 'binary'
            return int(G.has_edge(source, target))

    def product(path):
        effect = 1
        for i in range(len(path) - 1):
            effect *= sign(path[i], path[i + 1], prefix)
        return effect

    state_n = get_nodes(G, "state")
    input_n = get_nodes(G, "input")
    output_n = get_nodes(G, "output")
    matrix_configs = {
        "A": (state_n, state_n, "a", "state"),
        "B": (state_n, input_n, "b", "input"),
        "C": (output_n, state_n, "c", "output"),
        "D": (output_n, input_n, "d", "input"),
    }
    rows, cols, prefix, category = matrix_configs[matrix_type]
    matrix = sp.zeros(len(rows), len(cols))
    for i, target in enumerate(rows):
        for j, source in enumerate(cols):
            if matrix_type == "A":
                if G.has_edge(source, target):
                    matrix[i, j] = sign(source, target, prefix)
            else:
                paths = nx.all_simple_paths(G, source, target)
                valid = [
                    p
                    for p in paths
                    if all(G.nodes[n]["category"] == category for n in p[1:-1])
                ]
                matrix[i, j] = sum(product(path) for path in valid)
    return matrix


def create_equations(G, form="state", density_independent=None):
    if not isinstance(G, nx.DiGraph):
        raise TypeError("Input must be a networkx.DiGraph.")
    A = create_matrix(G, form="symbolic", matrix_type="A")
    B = create_matrix(G, form="symbolic", matrix_type="B")
    C = create_matrix(G, form="symbolic", matrix_type="C")
    D = create_matrix(G, form="symbolic", matrix_type="D")
    state_n = get_nodes(G, "state")
    input_n = get_nodes(G, "input")
    output_n = get_nodes(G, "output")
    x = sp.Matrix([sp.Symbol(f"x_{i}") for i in state_n])
    u = sp.Matrix([sp.Symbol(f"u_{i}") for i in input_n]) if input_n else None
    y = sp.Matrix([sp.Symbol(f"y_{i}") for i in output_n]) if output_n else None
    if density_independent is None:
        k = sp.zeros(A.shape[0], 1)
    else:
        if len(density_independent) != A.shape[0]:
            raise ValueError(f"The length of k must be equal to {A.shape[0]}")
        k = sp.Matrix(density_independent)
    state_equations = x.multiply_elementwise(A * x + k)
    if B.shape[1] > 0 and u is not None:
        state_equations += B * u
    output_equations = None
    if C.shape[0] > 0:
        output_equations = C * x
        if D.shape[1] > 0 and u is not None:
            output_equations += D * u
    if form == "vector":
        return A, B, C, D, x, u, y, k
    elif form == "state":
        return state_equations
    elif form == "output":
        if output_equations is None:
            print("No output equations available.")
        return output_equations
    elif form == "jacobian":
        return state_equations.jacobian(x)
    else:
        raise ValueError("Choose 'vector', 'state', 'output' or 'jacobian'.")



from .core.structure import (
    import_digraph,
    create_matrix,
    create_equations,
)

from .core.stability import (
    sign_stability,
    system_feedback,
    net_feedback,
    absolute_feedback,
    weighted_feedback,
    feedback_metrics,
    hurwitz_determinants,
    net_determinants,
    absolute_determinants,
    weighted_determinants,
    determinants_metrics,
    create_model_c,
    conditional_stability,
    simulation_stability,
)

from .core.press import (
    adjoint_matrix,
    absolute_feedback_matrix,
    weighted_predictions_matrix,
    sign_determinacy_matrix,
    numerical_simulations,
)

from .core.prediction import (
    table_of_predictions,
    compare_predictions,
    create_plot,
)

from .core.helper import (
    list_to_digraph,
    digraph_to_list,
    powerplay_labels,
    perm,
    get_nodes,
    get_positive,
    get_negative,
    get_weight,
    sign_determinacy,
)


from .extensions.sensitivity import (
    structural_sensitivity,
    net_structural_sensitivity,
    absolute_structural_sensitivity,
    weighted_structural_sensitivity,
)
from .extensions.life import (
    birth_matrix,
    death_matrix,
    life_expectancy_change,
    net_life_expectancy_change,
    absolute_life_expectancy_change,
    weighted_predictions_life_expectancy,
)
from .extensions.paths import (
    get_paths,
    paths_table,
    get_cycles,
    cycles_table,
    complementary_feedback,
    system_paths,
    weighted_paths,
)
from .extensions.effects import (
    define_state_space,
    cumulative_effects,
    net_effects,
    absolute_effects,
    positive_effects,
    negative_effects,
    weighted_effects,
    sign_determinacy_effects,
    get_simulations,
    simulation_effects,
)

from .extensions.indicators import mutual_information

from .extensions.validation import (
    marginal_likelihood,
    bayes_factors,
    posterior_predictions,
    diagnose_observations,
)


__all__ = [
    # structure.py
    "import_digraph",
    "create_matrix",
    "create_equations",
    # stability.py
    "sign_stability",
    "system_feedback",
    "net_feedback",
    "absolute_feedback",
    "weighted_feedback",
    "feedback_metrics",
    "hurwitz_determinants",
    "net_determinants",
    "absolute_determinants",
    "weighted_determinants",
    "determinants_metrics",
    "create_model_c",
    "conditional_stability",
    "simulation_stability",
    # press.py
    "adjoint_matrix",
    "absolute_feedback_matrix",
    "weighted_predictions_matrix",
    "sign_determinacy_matrix",
    "numerical_simulations",
    # prediction.py
    "table_of_predictions",
    "compare_predictions",
    "create_plot",
    # helper.py
    "list_to_digraph",
    "digraph_to_list",
    "powerplay_labels",
    "perm",
    "get_nodes",
    "get_positive",
    "get_negative",
    "get_weight",
    "sign_determinacy",
    # sensitivity.py
    "structural_sensitivity",
    "net_structural_sensitivity",
    "absolute_structural_sensitivity",
    "weighted_structural_sensitivity",
    # life.py
    "birth_matrix",
    "death_matrix",
    "life_expectancy_change",
    "net_life_expectancy_change",
    "absolute_life_expectancy_change",
    "weighted_predictions_life_expectancy",
    # paths.py
    "get_paths",
    "paths_table",
    "get_cycles",
    "cycles_table",
    "complementary_feedback",
    "system_paths",
    "weighted_paths",
    # effects.py
    "define_state_space",
    "cumulative_effects",
    "net_effects",
    "absolute_effects",
    "positive_effects",
    "negative_effects",
    "weighted_effects",
    "sign_determinacy_effects",
    "get_simulations",
    "simulation_effects",
    # indicators.py
    "mutual_information",
    # validation.py
    "marginal_likelihood",
    "bayes_factors",
    "posterior_predictions",
    "diagnose_observations",
    # helper.py
    "list_to_digraph",
    "digraph_to_list",
    "powerplay_labels",
    "perm",
    "get_nodes",
    "get_positive",
    "get_negative",
    "get_weight",
    "sign_determinacy",
]



import sympy as sp
import numpy as np
import pandas as pd
from functools import cache
from .effects import get_simulations
from ..core.helper import get_nodes


@cache
def marginal_likelihood(G, perturb, observe, n_sim=10000, distribution="uniform", seed=42):
    sims = get_simulations(G, n_sim, distribution, seed, perturb, observe)
    return sum(sims["valid_sims"]) / n_sim


@cache
def bayes_factors(G_list, perturb, observe, n_sim=10000, distribution="uniform", seed=42, names=None):
    likelihoods = [marginal_likelihood(G, perturb, observe, n_sim, distribution, seed) for G in G_list]
    model_names = names if names and len(names) == len(G_list) else [f"M_{i+1}" for i in range(len(G_list))]
    bayes_factors = {f"{model_names[i]}/{model_names[j]}": (
        float("inf") if likelihoods[j] == 0 and likelihoods[i] > 0 else
        0 if likelihoods[j] == 0 else likelihoods[i] / likelihoods[j]
    ) for i in range(len(G_list)) for j in range(i + 1, len(G_list))}

    return pd.DataFrame({
        "Model comparison": list(bayes_factors.keys()),
        "Likelihood 1": [likelihoods[i] for i in range(len(G_list)) for j in range(i + 1, len(G_list))],
        "Likelihood 2": [likelihoods[j] for i in range(len(G_list)) for j in range(i + 1, len(G_list))],
        "Bayes factor": list(bayes_factors.values()),
    })


@cache
def posterior_predictions(G, perturb, observe=None, n_sim=10000, dist="uniform", seed=42):
    sims = get_simulations(G, n_sim, dist, seed, perturb, observe)
    state_nodes, output_nodes = get_nodes(G, "state"), get_nodes(G, "output")
    n, m = len(state_nodes), len(output_nodes)
    valid_count = sum(sims["valid_sims"])
    tmat = sims["tmat"]
    if valid_count == 0:
        return sp.Matrix([np.nan] * (n + m))
    effects = np.array([e[:n+m] if len(e) >= n+m else np.pad(e, (0, n+m-len(e))) for e, v in zip(sims["effects"], sims["valid_sims"]) if v])
    positive = np.sum(effects > 0, axis=0)
    negative = np.sum(effects < 0, axis=0)
    smat = positive / valid_count
    tmat_np = np.array(tmat.tolist(), dtype=bool)
    perturb_index = sims["all_nodes"].index(perturb[0])
    smat = [np.nan if not tmat_np[i, perturb_index] else smat[i] for i in range(n + m)]
    if observe:
        for node, value in observe:
            index = state_nodes.index(node) if node in state_nodes else (n + output_nodes.index(node) if node in output_nodes else None)
            if index is not None:
                smat[index] = 1 if value > 0 else (0 if value < 0 else np.nan)
    
    smat = np.where(negative > positive, -negative / valid_count, smat)
    return sp.Matrix(smat)


@cache
def diagnose_observations(G, observe, n_sim=10000, distribution="uniform", seed=42):
    perturb_nodes = get_nodes(G, "state") + get_nodes(G, "input")
    results = []
    for node in perturb_nodes:
        for sign in [1, -1]:
            try:
                likelihood = marginal_likelihood(G, (node, sign), observe, n_sim, distribution, seed)
                results.append({"Perturbed node": node, "Perturbation sign": sign, "Marginal likelihood": likelihood})
            except Exception as e:
                print(f"Error for node {node} with sign {sign}: {str(e)}")
    
    return pd.DataFrame(results).sort_values("Marginal likelihood", ascending=False).reset_index(drop=True)




import numpy as np
import sympy as sp
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt


def table_of_predictions(M, t1=0.8, t2=1, index=None, columns=None):
    if isinstance(M, sp.Matrix):
        M = sp.matrix2numpy(M, dtype=float)
    conditions = [
        (lambda x: x == sp.nan, "0"),
        (lambda x: x >= t2, "+"),
        (lambda x: x >= t1 and x < t2, "(+)"),
        (lambda x: x > -t1 and x < t1, "?"),
        (lambda x: x > -t2 and x <= -t1, "(\u2212)"),
        (lambda x: x <= -t2, "\u2212"),
    ]

    predictions = np.empty(M.shape, dtype=object)
    for i in range(M.shape[0]):
        for j in range(M.shape[1]):
            value = M[i, j]
            predictions[i, j] = next(
                (val for cond, val in conditions if cond(value)), "0"
            )
    return pd.DataFrame(predictions, index=index, columns=columns)


def compare_predictions(M1, M2):
    if not M1.index.equals(M2.index) or not M1.columns.equals(M2.columns):
        raise ValueError("M1 and M2 must have the same index and columns")
    combined = np.vectorize(lambda x, y: x if x == y else f"{x}, {y}")(
        M1.values, M2.values
    )
    return pd.DataFrame(combined, index=M1.index, columns=M1.columns)


def create_plot(data, **kwargs):
    plt.rcParams.update(
        {
            "xtick.top": True,
            "xtick.bottom": False,
            "xtick.labeltop": True,
            "xtick.labelbottom": False,
            "xtick.major.width": 0.5,
            "ytick.major.width": 0.5,
            "xtick.major.size": 3,
            "ytick.major.size": 3,
            "xtick.minor.size": 1.5,
            "ytick.minor.size": 1.5,
        }
    )
    args = {
        "annot": True,
        "linewidths": 0.75,
        "linecolor": "white",
        "cbar": False,
        "cmap": None,
    }
    args.update(kwargs)
    figsize = args.pop("figsize", None)
    if figsize:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig, ax = plt.subplots()
    sns.heatmap(data, ax=ax, **args)
    plt.setp(ax.get_xticklabels(), rotation=0, ha="center")
    plt.setp(ax.get_yticklabels(), rotation=0, ha="right")
    for spine in ax.spines.values():
        spine.set_visible(True)
        spine.set_color("white")
        spine.set_linewidth(0.5)
    return fig, ax



import pandas as pd
import numpy as np
import networkx as nx
from functools import cache
from ..core.helper import get_nodes
from .effects import get_simulations


@cache
def mutual_information(models, inputs, n_sim=10000, seed=42):
    models = [models] if not isinstance(models, (list, tuple)) else list(models)
    models = [nx.DiGraph(G) if not isinstance(G, nx.DiGraph) else G for G in models]
    inputs = (inputs,) if isinstance(inputs[0], str) else inputs
    if not all(isinstance(i, tuple) for i in inputs):
        raise ValueError("inputs must be a tuple or a tuple of tuples")
    nodes = sorted(set(node for G in models for node in get_nodes(G, "state") + get_nodes(G, "output")))
    all_effects = []
    for G in models:
        for perturb in inputs:
            sims = get_simulations(G, n_sim=n_sim, seed=seed, perturb=perturb)
            response_nodes = get_nodes(G, "state") + get_nodes(G, "output")
            node_map = {node: i for i, node in enumerate(response_nodes)}
            sim_effects = []
            for effect in sims["effects"]:
                sim_effects.append([effect[node_map[n]] if n in node_map and node_map[n] < len(effect) else np.nan for n in nodes])
            all_effects.append(np.array(sim_effects))
    mi_vals = []
    for i, node in enumerate(nodes):
        node_effects = [effects[:, i] for effects in all_effects]
        if any(np.all(np.isnan(effects)) for effects in node_effects):
            mi_vals.append(0)
            continue
        node_effects = np.concatenate(node_effects)
        labels = np.concatenate([np.full(effects.shape[0], i) for i, effects in enumerate(all_effects)])
        valid = ~np.isnan(node_effects)
        node_effects, labels = node_effects[valid], labels[valid]
        if len(node_effects) == 0:
            mi_vals.append(0)
            continue
        joint, _, _ = np.histogram2d(labels, node_effects > 0, bins=(len(models) * len(inputs), 2))
        joint_p = joint / joint.sum()
        l_p, e_p = joint_p.sum(axis=1), joint_p.sum(axis=0)
        mi = sum(joint_p[i, j] * np.log2(joint_p[i, j] / (l_p[i] * e_p[j] + 1e-10))
                 for i in range(len(models) * len(inputs))
                 for j in range(2)
                 if joint_p[i, j] > 0)
        mi_vals.append(max(0, mi))
    return pd.DataFrame({"Node": nodes, "Mutual Information": mi_vals}).sort_values("Mutual Information", ascending=False).reset_index(drop=True)



import numpy as np
import sympy as sp
import networkx as nx
from functools import cache
from ..core.helper import get_nodes, get_weight, get_positive, get_negative, sign_determinacy
from ..core.structure import create_matrix
from ..core.press import adjoint_matrix, absolute_feedback_matrix

def define_state_space(G, remove_disconnected=True):
    if not isinstance(G, nx.DiGraph):
        raise TypeError("Input must be a networkx.DiGraph.")

    G_def = G.copy()

    if remove_disconnected:
        G_undirected = G_def.to_undirected()
        connected = list(nx.connected_components(G_undirected))
        if len(connected) > 1:
            largest = max(connected, key=len)
            nodes_to_remove = [
                node for system in connected if system != largest for node in system
            ]
            G_def.remove_nodes_from(nodes_to_remove)

    nx.set_node_attributes(G_def, "state", "category")

    while True:
        reclassified = False
        for node in list(G_def.nodes()):
            if G_def.nodes[node]["category"] == "state":
                if all(
                    G_def.nodes[pred]["category"] == "input"
                    for pred in G_def.predecessors(node)
                ):
                    G_def.nodes[node]["category"] = "input"
                    reclassified = True
                elif all(
                    G_def.nodes[succ]["category"] == "output"
                    for succ in G_def.successors(node)
                ):
                    G_def.nodes[node]["category"] = "output"
                    reclassified = True
        if not reclassified:
            break

    return G_def


@cache
def cumulative_effects(G, form="symbolic"):
    B = create_matrix(G, form=form, matrix_type="B")
    C = create_matrix(G, form=form, matrix_type="C")
    D = create_matrix(G, form=form, matrix_type="D")
    if form == "symbolic":
        effects = adjoint_matrix(G, form="symbolic")
    elif form in "signed":
        effects = adjoint_matrix(G, form="signed")
    elif form in "binary":
        effects = absolute_feedback_matrix(G)
    else:
        raise ValueError("Invalid form. Choose 'symbolic', 'signed', 'binary'.")
    cemat = sp.BlockMatrix(
        [[effects, effects * B], [C * effects, C * effects * B + D]]
    ).as_explicit()
    if form != "symbolic":
        cemat = cemat.subs({sym: 1 for sym in cemat.free_symbols})
    return sp.expand(cemat)


@cache
def net_effects(G):
    return cumulative_effects(G, form="signed")


@cache
def absolute_effects(G):
    return cumulative_effects(G, form="binary")


@cache
def positive_effects(G):
    net = net_effects(G)
    absolute = absolute_effects(G)
    return get_positive(net, absolute)


@cache
def negative_effects(G):
    net = net_effects(G)
    absolute = absolute_effects(G)
    return get_negative(net, absolute)


@cache
def weighted_effects(G):
    net = net_effects(G)
    absolute = absolute_effects(G)
    return get_weight(net, absolute)


@cache
def sign_determinacy_effects(G, method="average"):
    wmat = weighted_effects(G)
    tmat = absolute_effects(G)
    return sign_determinacy(wmat, tmat, method=method)

@cache
def get_simulations(G, n_sim=10000, dist="uniform", seed=42, perturb=None, observe=None):
    np.random.seed(seed)
    A, B, C, D = [create_matrix(G, form="symbolic", matrix_type=m) for m in "ABCD"]
    state_nodes, input_nodes, output_nodes = [get_nodes(G, t) for t in ["state", "input", "output"]]
    all_nodes = state_nodes + input_nodes + output_nodes
    node_idx = {node: i for i, node in enumerate(all_nodes)}
    tmat = sp.matrix2numpy(absolute_effects(G)).astype(int)
    dist_funcs = {
        "uniform": lambda size: np.random.uniform(0, 1, size),
        "weak": lambda size: np.random.beta(1, 3, size),
        "moderate": lambda size: np.random.beta(2, 2, size),
        "strong": lambda size: np.random.beta(3, 1, size)
    }
    pert_idx, perturb_sign = (node_idx[perturb[0]], perturb[1]) if perturb else (None, 1)
    n_state, n_input, n_output = len(state_nodes), len(input_nodes), len(output_nodes)
    n_rows, n_cols = n_state + n_output, n_state + n_input
    symbols = list(set(A.free_symbols) | set(B.free_symbols) | set(C.free_symbols) | set(D.free_symbols))
    A_sp = sp.lambdify(symbols, A)
    B_sp = sp.lambdify(symbols, B) if n_input > 0 else None
    C_sp = sp.lambdify(symbols, C) if n_output > 0 else None
    D_sp = sp.lambdify(symbols, D) if D.shape != (0, 0) else None
    effects, valid_sims = [], []
    for _ in range(n_sim * 100):
        values = dist_funcs[dist](len(symbols))
        sim_A = A_sp(*values)
        if np.all(np.real(np.linalg.eigvals(sim_A)) < 0):
            try:
                inv_A = np.linalg.inv(-sim_A)
                B_np = B_sp(*values) if B_sp else np.array([]).reshape(n_state, 0)
                D_np = D_sp(*values) if D_sp else np.array([]).reshape(n_output, n_input)
                C_np = C_sp(*values) if C_sp else np.array([]).reshape(0, n_state)
                effect_matrix = np.zeros((n_rows, n_cols))
                if n_state > 0:
                    effect_matrix[:n_state, :n_state] = inv_A
                    if n_input > 0:
                        effect_matrix[:n_state, n_state:] = inv_A @ B_np
                if n_output > 0:
                    if n_state > 0:
                        effect_matrix[n_state:, :n_state] = C_np @ inv_A
                        if n_input > 0:
                            effect_matrix[n_state:, n_state:] = C_np @ inv_A @ B_np + D_np
                    elif n_input > 0:
                        effect_matrix[n_state:, n_state:] = D_np
                effect = effect_matrix[:, pert_idx] * perturb_sign if pert_idx is not None else effect_matrix
                effects.append(effect)
                if observe:
                    valid = all((node in state_nodes and ((tmat[node_idx[node], pert_idx] == 0 and obs == 0) or
                                 (obs != 0 and np.sign(effect[node_idx[node]]) == obs))) or
                                (node in output_nodes and ((tmat[len(state_nodes) + output_nodes.index(node), pert_idx] == 0
                                 and obs == 0) or (obs != 0 and np.sign(effect[len(state_nodes) + output_nodes.index(node)]) == obs)))
                                for node, obs in observe)
                    valid_sims.append(valid)
                else:
                    valid_sims.append(True)
                if len(effects) == n_sim:
                    break
            except np.linalg.LinAlgError:
                continue
    else:
        raise RuntimeError(f"Maximum iterations reached. Stable proportion: {len(effects) / (n_sim * 100):.4f}")
    return {"effects": effects, "valid_sims": valid_sims, "all_nodes": all_nodes, "tmat": tmat}


def simulation_effects(G, n_sim=10000, dist="uniform", seed=42):
    sims = get_simulations(G, n_sim, dist, seed)
    tmat = sims["tmat"]
    state_nodes = get_nodes(G, "state")
    input_nodes = get_nodes(G, "input")
    output_nodes = get_nodes(G, "output")
    n_state, n_input, n_output = len(state_nodes), len(input_nodes), len(output_nodes)
    n_rows, n_cols = n_state + n_output, n_state + n_input
    positive = np.zeros((n_rows, n_cols), dtype=int)
    negative = np.zeros((n_rows, n_cols), dtype=int)
    for effect in sims["effects"]:
        positive += effect > 0
        negative += effect < 0
    smat = np.where(negative > positive, -negative / n_sim, positive / n_sim)
    smat = sp.Matrix(smat)
    smat = sp.Matrix([[sp.nan if not tmat[i, j] else smat[i, j] for j in range(smat.cols)]
                      for i in range(smat.rows)])
    return smat
